@@ -29,99 +29,81 @@ sub new {
     return $self;
 }
 
+# (recursively) checks data, or a section thereof;
+# by instantiatin D::P::V objects and calling validate on them
+
 sub validate {
     my $self = shift;
     $self->{errors} = Data::Processor::Error::Collection->new();
-    $self->_validate($self->{data}, $self->{schema}, 'root');
-    return $self->{errors};
-}
 
-#################
-# internal methods
-#################
-sub _validate {
-    my $self = shift;
-    # $(word)_section are *not* the data fields but the sections of the
-    # data / schema the recursive algorithm is currently working on.
-    # (Only) in the first call, these are identical.
-    my $data_section = shift;
-    my $schema_section = shift;
+    $self->_add_defaults();
 
-    $self->_add_defaults($data_section, $schema_section);
-
-    for my $key (keys %{$data_section}){
+    for my $key (keys %{$self->{data}}){
         $self->explain (">>'$key'");
 
-         # checks
-        my $key_schema_to_descend_into =
-            $self->__key_present_in_schema(
-                $key, $data_section, $schema_section
-            );
-
-        $self->__value_is_valid(
-            $key, $data_section, $schema_section
-        );
-
-        $self->__validator_returns_undef(
-            $key, $data_section, $schema_section
-        ) if exists $schema_section->{$key}
-             and exists $schema_section->{$key}->{validator};
+        # checks
+        my $schema_key = $self->_schema_twin_key($key) or next;
+        # from here we know to have a "twin" key $schema_key in the schema
+        $self->__value_is_valid( $key );
+        $self->__validator_returns_undef($key, $schema_key);
 
         # transformer
-        my $e = $self->{transformer}
-                ->transform($schema_section, $data_section, $key);
+        my $e = $self->{transformer}->transform($key, $self);
         $self->error($e) if $e;
 
-        my $descend_into;
-        if (exists  $schema_section->{$key}
-                and $schema_section->{$key}->{no_descend_into}
-                and $schema_section->{$key}->{no_descend_into}){
+        # skip if explicitly asked for
+        if ($self->{schema}->{$schema_key}->{no_descend_into}){
             $self->explain (
                 "skipping '$key' because schema explicitly says so.\n");
+            next;
         }
         # skip data branch if schema key is empty.
-        elsif (exists $schema_section->{$key}
-                and ! %{$schema_section->{$key}}){
-            $self->explain (
-                "skipping '$key' because schema key is empty'");
+        if (! %{$self->{schema}->{$schema_key}}){
+            $self->explain ("skipping '$key' because schema key is empty'");
+            next;
         }
-        elsif (exists $schema_section->{$key}
-                and ! exists $schema_section->{$key}->{members}){
+        if (! $self->{schema}->{$schema_key}->{members}){
             $self->explain (
                 "not descending into '$key'. No members specified\n"
             );
-        }
-        else{
-            $descend_into = 1;
-            $self->explain (">>descending into '$key'\n");
+            next;
         }
 
-        # recursion
-        if ((ref $data_section->{$key} eq ref {})
-                and $descend_into){
-            $self->explain (">>'$key' is not a leaf and we descend into it\n");
-            push @{$self->{parent_keys}}, $key;
-            $self->{depth}++;
-            $self->_validate(
-                $data_section->{$key},
-                $schema_section->{$key_schema_to_descend_into}->{members}
-            );
-            # to undo push before entering recursion.
-            pop @{$self->{parent_keys}};
-            $self->{depth}--;
-        }
-        elsif ((ref $data_section->{$key} eq ref []) && $descend_into
-            && exists $schema_section->{$key_schema_to_descend_into}->{array}
-            && $schema_section->{$key_schema_to_descend_into}->{array}){
+        # recursion if we reach this point.
+        $self->explain (">>descending into '$key'\n");
 
-            $self->explain(">>'$key' is an array reference so we check all elements\n");
+        if (ref $self->{data}->{$key} eq ref {} ){
+            $self->explain
+                (">>'$key' is not a leaf and we descend into it\n");
+            my $e = Data::Processor::Validator->new(
+                schema      => $self->{schema}->{$schema_key}->{members},
+                data        => $self->{data}->{$key},
+                parent_keys => [@{$self->{parent_keys}}, $key],
+                depth       => $self->{depth}+1,
+                verbose     => $self->{verbose},
+
+            ) ->validate();
+            $self->{errors}->add_collection($e);
+
+        }
+        elsif ((ref $self->{data}->{$key} eq ref [])
+            && $self->{schema}->{$schema_key}->{array}){
+
+            $self->explain(
+            ">>'$key' is an array reference so we check all elements\n");
             push @{$self->{parent_keys}}, $key;
             $self->{depth}++;
-            for my $member (@{$data_section->{$key}}){
-                $self->_validate(
-                    $member,
-                    $schema_section->{$key_schema_to_descend_into}->{members}
-                );
+            for my $member (@{$self->{data}->{$key}}){
+                my $e = Data::Processor::Validator->new(
+                    schema      => $self->{schema}->{$schema_key}->{members},
+                    data        => $member,
+                    parent_keys => [@{$self->{parent_keys}}, $key],
+                    depth       => $self->{depth}+1,
+                    verbose     => $self->{verbose},
+
+                ) ->validate();
+                $self->{errors}->add_collection($e);
+
             }
             pop @{$self->{parent_keys}};
             $self->{depth}--;
@@ -131,14 +113,7 @@ sub _validate {
         # but it might be required by the schema.
         else {
             $self->explain(">>checking data key '$key' which is a leaf..");
-            if ( $key_schema_to_descend_into
-                    and
-                 $schema_section->{$key_schema_to_descend_into}
-                    and
-                ref $schema_section->{$key_schema_to_descend_into} eq ref {}
-                    and
-                exists $schema_section->{$key_schema_to_descend_into}->{members}
-            ){
+            if ($self->{schema}->{$schema_key}->{members}){
                 $self->explain("but schema requires members.\n");
                 $self->error("'$key' should have members");
             }
@@ -146,18 +121,18 @@ sub _validate {
                 $self->explain("schema key is also a leaf. ok.\n");
             }
         }
-
-
     }
-     # look for missing non-optional keys in schema
+    # look for missing non-optional keys in schema
     # this is only done on this level.
     # Otherwise "mandatory" inherited "upwards".
-    $self->_check_mandatory_keys(
-        $data_section, $schema_section
-    );
-
+    $self->_check_mandatory_keys();
+    return $self->{errors};
 }
 
+#################
+# internal methods
+#################
+
 # add an error
 sub error {
     my $self = shift;
@@ -183,14 +158,12 @@ sub explain {
 # default value.
 
 sub _add_defaults{
-    my $self           = shift;
-    my $data_section = shift;
-    my $schema_section = shift;
+    my $self    = shift;
 
-    for my $key (keys %{$schema_section}){
-        next unless exists $schema_section->{$key}->{default};
-        $data_section->{$key} = $schema_section->{$key}->{default}
-            unless $data_section->{$key};
+    for my $key (keys %{$self->{schema}}){
+        next unless $self->{schema}->{$key}->{default};
+        $self->{data}->{$key} = $self->{schema}->{$key}->{default}
+            unless $self->{data}->{$key};
     }
 }
 
@@ -198,40 +171,36 @@ sub _add_defaults{
 # below current level (in schema)
 # for each check if $data has a key.
 sub _check_mandatory_keys{
-    my $self = shift;
-    my $data_section = shift;
-    my $schema_section = shift;
+    my $self    = shift;
 
-    for my $key (keys %{$schema_section}){
+    for my $key (keys %{$self->{schema}}){
         $self->explain(">>Checking if '$key' is mandatory: ");
-        unless (exists $schema_section->{$key}->{optional}
-                   and $schema_section->{$key}->{optional}){
+        unless ($self->{schema}->{$key}->{optional}
+                   and $self->{schema}->{$key}->{optional}){
 
             $self->explain("true\n");
-            next if exists $data_section->{$key};
+            next if $self->{data}->{$key};
 
             # regex-keys never directly occur.
-            if (exists $schema_section->{$key}->{regex}
-                   and $schema_section->{$key}->{regex}){
+            if ($self->{schema}->{$key}->{regex}){
                 $self->explain(">>regex enabled key found. ");
                 $self->explain("Checking data keys.. ");
                 my $c = 0;
                 # look which keys match the regex
-                for my $c_key (keys %{$data_section}){
+                for my $c_key (keys %{$self->{data}}){
                     $c++ if $c_key =~ /$key/;
                 }
                 $self->explain("$c matching occurencies found\n");
                 next if $c > 0;
             }
-            next if exists $schema_section->{$key}->{array}
-                && $schema_section->{$key}->{array};
+            next if $self->{schema}->{$key}->{array};
 
 
             # should only get here in case of error.
 
             my $error_msg = '';
-            $error_msg = $schema_section->{$key}->{error_msg}
-                if exists $schema_section->{$key}->{error_msg};
+            $error_msg = $self->{schema}->{$key}->{error_msg}
+                if $self->{schema}->{$key}->{error_msg};
             $self->error("mandatory key '$key' missing. Error msg: '$error_msg'");
         }
         else{
@@ -240,63 +209,60 @@ sub _check_mandatory_keys{
     }
 }
 
-# called by _validate to check if a given key is defined in schema
-sub __key_present_in_schema{
-    my $self = shift;
-    my $key            = shift;
-    my $data_section = shift;
-    my $schema_section = shift;
+# called by validate to check if a given key is defined in schema
+sub _schema_twin_key{
+    my $self    = shift;
+    my $key     = shift;
 
-    my $key_schema_to_descend_into;
+    my $schema_key;
 
     # direct match: exact declaration
-    if (exists $schema_section->{$key}){
+    if ($self->{schema}->{$key}){
         $self->explain(" ok\n");
-        $key_schema_to_descend_into = $key;
+        $schema_key = $key;
     }
     # match against a pattern
     else {
         my $match;
-        for my $match_key (keys %{$schema_section}){
+        for my $match_key (keys %{$self->{schema}}){
 
             # only try to match a key if it has the property
             # _regex_ set
-            next unless exists $schema_section->{$match_key}
-                    and exists $schema_section->{$match_key}->{regex}
-                           and $schema_section->{$match_key}->{regex};
+            next unless exists $self->{schema}->{$match_key}
+                           and $self->{schema}->{$match_key}->{regex};
 
             if ($key =~ /$match_key/){
                 $self->explain("'$key' matches $match_key\n");
-                $key_schema_to_descend_into = $match_key;
+                $schema_key = $match_key;
             }
         }
     }
 
-    # if $key_schema_to_descend_into is still undef we were unable to
+    # if $schema_key is still undef we were unable to
     # match it against a key in the schema.
-    unless ($key_schema_to_descend_into){
+    unless ($schema_key){
         $self->explain(">>$key not in schema, keys available: ");
-        $self->explain(join (", ", (keys %{$schema_section})));
+        $self->explain(join (", ", (keys %{$self->{schema}})));
         $self->explain("\n");
         $self->error("key '$key' not found in schema\n");
     }
-    return $key_schema_to_descend_into
+    return $schema_key
 }
 
 # 'validator' specified gets this called to call the callback :-)
 sub __validator_returns_undef {
-    my $self = shift;
-    my $key    = shift;
-    my $data_section = shift;
-    my $schema_section = shift;
-    $self->explain("running validator for '$key': $data_section->{$key}\n");
+    my $self       = shift;
+    my $key        = shift;
+    my $schema_key = shift;
+    return unless $self->{schema}->{$schema_key}->{validator};
+    $self->explain("running validator for '$key': $self->{data}->{$key}\n");
 
-    if (ref $data_section->{$key} eq ref []
-        && exists $schema_section->{$key}->{array} && $schema_section->{$key}->{array}){
+    if (ref $self->{data}->{$key} eq ref []
+        && $self->{schema}->{$key}->{array}){
 
         my $counter = 0;
-        for my $elem (@{$data_section->{$key}}){
-            my $return_value = $schema_section->{$key}->{validator}->($elem, $data_section);
+        for my $elem (@{$self->{data}->{$key}}){
+            my $return_value = $self->{schema}->{$key}->{validator}->($elem, $self->{data});
             if ($return_value){
                 $self->explain("validator error: $return_value (element $counter)\n");
                 $self->error("Execution of validator for '$key' element $counter returns with error: $return_value");
@@ -308,7 +274,7 @@ sub __validator_returns_undef {
         }
     }
     else {
-        my $return_value = $schema_section->{$key}->{validator}->($data_section->{$key}, $data_section);
+        my $return_value = $self->{schema}->{$key}->{validator}->($self->{data}->{$key}, $self->{data});
         if ($return_value){
             $self->explain("validator error: $return_value\n");
             $self->error("Execution of validator for '$key' returns with error: $return_value");
@@ -319,51 +285,49 @@ sub __validator_returns_undef {
     }
 }
 
-# called by _validate to check if a value is in line with definitions
+# called by validate to check if a value is in line with definitions
 # in the schema.
 sub __value_is_valid{
-    my $self = shift;
-    my $key    = shift;
-    my $data_section = shift;
-    my $schema_section = shift;
+    my $self    = shift;
+    my $key     = shift;
 
-    if (exists  $schema_section->{$key}
-            and $schema_section->{$key}->{value}){
-        $self->explain('>>'.ref($schema_section->{$key}->{value})."\n");
+    if (exists  $self->{schema}->{$key}
+            and $self->{schema}->{$key}->{value}){
+        $self->explain('>>'.ref($self->{schema}->{$key}->{value})."\n");
 
         # currently, 2 type of restrictions are supported:
         # (callback) code and regex
-        if (ref($schema_section->{$key}->{value}) eq 'CODE'){
+        if (ref($self->{schema}->{$key}->{value}) eq 'CODE'){
             # possibly never implement this because of new "validator"
         }
-        elsif (ref($schema_section->{$key}->{value}) eq 'Regexp'){
-            if (ref $data_section->{$key} eq ref []
-                && exists $schema_section->{$key}->{array} && $schema_section->{$key}->{array}){
+        elsif (ref($self->{schema}->{$key}->{value}) eq 'Regexp'){
+            if (ref $self->{data}->{$key} eq ref []
+                && $self->{schema}->{$key}->{array}){
 
-                for my $elem (@{$data_section->{$key}}){
-                    $self->explain(">>match '$elem' against '$schema_section->{$key}->{value}'");
+                for my $elem (@{$self->{data}->{$key}}){
+                    $self->explain(">>match '$elem' against '$self->{schema}->{$key}->{value}'");
 
-                    if ($elem =~ m/^$schema_section->{$key}->{value}$/){
+                    if ($elem =~ m/^$self->{schema}->{$key}->{value}$/){
                         $self->explain(" ok.\n");
                     }
                     else{
                         # XXX never reach this?
                         $self->explain(" no.\n");
-                        $self->error("$elem does not match ^$schema_section->{$key}->{value}\$");
+                        $self->error("$elem does not match ^$self->{schema}->{$key}->{value}\$");
                     }
                 }
             }
             # XXX this was introduced to support arrays.
             else {
-               $self->explain(">>match '$data_section->{$key}' against '$schema_section->{$key}->{value}'");
+               $self->explain(">>match '$self->{data}->{$key}' against '$self->{schema}->{$key}->{value}'");
 
-                if ($data_section->{$key} =~ m/^$schema_section->{$key}->{value}$/){
+                if ($self->{data}->{$key} =~ m/^$self->{schema}->{$key}->{value}$/){
                     $self->explain(" ok.\n");
                 }
                 else{
                     # XXX never reach this?
                     $self->explain(" no.\n");
-                    $self->error("$data_section->{$key} does not match ^$schema_section->{$key}->{value}\$");
+                    $self->error("$self->{data}->{$key} does not match ^$self->{schema}->{$key}->{value}\$");
                 }
             }
         }
